package com.urbancode.anthill3.plugin.scm

import com.urbancode.anthill3.*
import com.urbancode.anthill3.plugin.scm.changelog.*;

import java.util.Date;
import java.util.TimeZone;
import java.text.SimpleDateFormat


public class SCMChangelog extends SCMStep {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    // changelog boundaries
    Date startDate;
    Date endDate;
    String startRevision;
    String endRevision;

    // filters for changeset
    final ChangeSetFilter changeSetFilter = new ChangeSetFilter();

    /**
     *
     * @return xml document of changes
     */
    public def execute() {
        //
        // Authenticate if not currently logged in
        //
        
        this.login();
        
        //
        // Get Change Log
        //

        def logOutput = runLogCommand()

        //
        // Parse Change Log into changesets
        //

        // the result change set data
        List<ChangeSet> changeSets = parseLogIntoChangeSets(logOutput);

        //
        // Construct AHP XML message and upload to AHP
        //

        if (changeSets) {
            ChangeSetXMLHelper xmlHelper = new ChangeSetXMLHelper()
            return xmlHelper.toXML(changeSets);
        }
        else {
            println "No changes detected"
            return null
        }
    }

    /**
     * Parse an input of accurev xml hist command into a list of AHP ChangeSets
     * @param logOutput output from a `accurev hist -fx` command
     * @return list of ahp changesets
     */
    public List<ChangeSet> parseLogIntoChangeSets(String logOutput) {

        List<ChangeSet> changeSets = []

        groovy.util.slurpersupport.GPathResult log = new XmlSlurper().parseText(logOutput);

        log.transaction.each{ trans ->
            ChangeSet changeSet = new ChangeSet();
            changeSet.id = trans.@id.text();
            changeSet.user = trans.@user.text();
            changeSet.date = new Date((trans.@time.text() as Long) * 1000);
            changeSet.message = trans.comment.text();

            def changeSetType = trans.@type.text()
            def workItems = new LinkedHashSet();

            trans.version.each{ ver ->
                def path = ver.@path.text() ;

                // Accurev seems to prefix all of their file paths with
                // /./ so I am removing it here for consistency with the rest
                // of our file paths.
                path = path.replace('^/\\./', '');

                ChangeSetFileEntry entry = new ChangeSetFileEntry();
                entry.revision = changeSet.id;
                entry.type = changeSetType;
                entry.path = path;
                changeSet.fileSet << entry;

                // find workItem meta-data elements
                ver.issueNum.each{ workItems << it.text() }
            }

            if (workItems) {
                changeSet.properties['workItems'] = workItems.join(',');
            }

            boolean allowAuthor = changeSet.hasAllowedAuthor(changeSetFilter);
            boolean allowAnyPaths = changeSet.hasAllowedPath(changeSetFilter);
            if (allowAuthor && allowAnyPaths) {
                changeSets << changeSet
            }
            else {
                def message = new StringBuilder("Changeset ${changeSet.id} skipped because ")
                if (!allowAuthor) {
                    message << "it has excluded author ${changeSet.user}"
                }
                if (!allowAuthor && !allowAnyPaths) {
                    message << " and "
                }
                if (!allowAnyPaths) {
                    message << "it contains only excluded file paths (${changeSet.fileSet.collect{it.path}})"
                }
                println message
            }
        }

        return changeSets;
    }

    protected String calculateRange() {

        // Should range use to startBuildlifeId["workspace.revision.$srcName"] ||
        //       blProps["workspace.revision.$srcName"] if present ?

        final def range;
        if (startRevision && endRevision) {
            range = "$endRevision-$startRevision"
        }
        else if (startDate && endDate) {
            range = scmHelper.formatDateRange(startDate, endDate)
        }
        else {
            range = "";
        }
        return range;
    }

    protected String runLogCommand() {
        final def cmdHelper = new CommandHelper(directory);
        def logCommand = [this.scmCommand, "hist", '-a', '-s', streamName, '-fx']
        
        final def range = calculateRange();
        if (range) {
            logCommand << '-t' << range
            if (host) {
                logCommand << '-H' << host
            }
            
            def logOutput = null
            cmdHelper.runCommand('Getting Changelog', logCommand) { Process proc ->
                proc.out.close()
                proc.consumeProcessErrorStream(System.out)
                logOutput = proc.text
                println logOutput
                if(!(logOutput?.trim())) {
                    println("No Changes Found");
                }
            }
            return logOutput;
        }
        else {
            println "WARNING: No start point specified, skipping changelog"
            return "";
        }
    }
    
}
